#  ---------------------------------------------------------------------
#   Metadata Editor version 1.0
#    v1.0
#    April 2002
#  ------------------+--------------------+-----------------------------
#   Hussein Suleman  |   hussein@vt.edu   |    www.husseinsspace.com    
#  ------------------+--------------------+-+---------------------------
#   Department of Computer Science          |        www.cs.vt.edu       
#     Digital Library Research Laboratory   |       www.dlib.vt.edu      
#  -----------------------------------------+-------------+-------------
#   Virginia Polytechnic Institute and State University   |  www.vt.edu  
#  -------------------------------------------------------+-------------

package Pure::MDEdit;


use Pure::EZXML;
use Pure::FileBox;


# constructor to read in parameters and schema and normalize schema types
sub new
{
   my ($classname, $schemafile, $options) = @_;
   my $self = {
      plusurl      => 'http://oai.dlib.vt.edu/~hussein/plus.gif',
      delurl       => 'http://oai.dlib.vt.edu/~hussein/del.gif',
      missingcolor => '#eeee00',
      cgideco      => 0,
      mdtemp       => 'mdtemp',
      filebox      => new Pure::FileBox ({}),
   };
   bless $self, $classname;
   
   my $parser = new Pure::EZXML;
   
   my $schema = $parser->parsefile ($schemafile)->getDocumentElement;

   # add in external lists
   if (defined $options)
   {
      if (exists $options->{external})
      {
         foreach my $key (keys %{$options->{external}})
         {
            foreach my $externallist ($schema->getElementsByTagName ('externallist'))
            {
               if ($externallist->getFirstChild->getNodeValue eq $key)
               {
                  my $element = $externallist->getParentNode->getParentNode->getParentNode;
                  my $restriction = $element->getElementsByTagName ('simpleType', 0)->item(0)
                                            ->getElementsByTagName ('restriction', 0)->item(0);
                  foreach my $value (@{$options->{external}->{$key}})
                  {
                     my $frag = $parser->parse ("<enumeration value=\"$value\"/>")->getDocumentElement;
                     $restriction->appendChild ($frag);
                  }
               }
            }
         }
      }
      if (exists $options->{schemaURL})
      {
         $self->{schemaURL} = $options->{schemaURL};
      }
      if (exists $options->{mdtemp})
      {
         $self->{mdtemp} = $options->{mdtemp};
      }
      if (exists $options->{filebox})
      {
         $self->{filebox} = new Pure::FileBox ($options->{filebox});
      }
   }
   
   $self->normalize ($schema);
   $self->{schema} = $schema->getElementsByTagName ('element', 0)->item(0);
   
   return $self;
}


# destructor
sub dispose 
{
   my ($self) = @_;
}


# replicate type information for every element of that type
sub flattenElement
{
   my ($self, $element, $document) = @_;
   
   my $type = $element->getAttribute ('type');
   
   # ignore namespace
   $type =~ s/^([^:]+:)?(.*)$/$2/;
   
   if ($type ne 'string')
   {
      foreach my $complexType ($document->getElementsByTagName ('complexType', 0))
      {
         if ($complexType->getAttribute ('name') eq $type)
         {
            foreach my $sequence ($complexType->getElementsByTagName ('sequence', 0))
            {
               foreach my $innerelement ($sequence->getElementsByTagName ('element', 0))
               {
                  $self->flattenElement ($innerelement, $document);
               }
            }
            my $clone = $complexType->cloneNode ('true');
            $element->appendChild ($clone);
            $clone->removeAttribute ('name');
            $element->removeAttribute ('type');
            return;
         }
      }
      foreach my $simpleType ($document->getElementsByTagName ('simpleType', 0))
      {
         if ($simpleType->getAttribute ('name') eq $type)
         {
            my $clone = $simpleType->cloneNode ('true');
            $element->appendChild ($clone);
            $clone->removeAttribute ('name');
            $element->removeAttribute ('type');
            return;
         }
      }
   }
}


# normalize schema by removing filtering global types into the places they are used
sub normalize
{
   my ($self, $document) = @_;
   
   foreach my $element ($document->getElementsByTagName ('element', 0))
   {
      $self->flattenElement ($element, $document);
   }

   foreach my $element ($document->getElementsByTagName ('complexType', 0))
   {
      $document->removeChild ($element);
   }
   foreach my $element ($document->getElementsByTagName ('simpleType', 0))
   {
      $document->removeChild ($element);
   }
}


# merge XML source document with schema
sub decorateFromXML
{
   my ($self, $schema, $instance) = @_;
 
   if (($schema->getAttribute ('type') eq 'string') ||
       ($schema->getElementsByTagName ('simpleType', 0)->getLength == 1))
   {
      $instance->setTagName ('value');
      $schema->appendChild ($instance);
   }
   else
   {
      my $template = $schema->getElementsByTagName ('complexType', 0)->item(0);
      my $clone = $template->cloneNode ('true');
      $clone->setTagName ('value');
      $schema->insertBefore ($clone, $template);
      foreach my $instancechild ($instance->getChildNodes)
      {
         if ($instancechild->getNodeType == ELEMENT_NODE)
         {
            my @possibles = $clone->getElementsByTagName ('sequence', 0)->item(0)
                                  ->getElementsByTagName ('element', 0);
            @possibles = grep {
               $_->getAttribute ('name') eq $instancechild->getNodeName
            } @possibles;
            if ($#possibles >= 0)
            {
               $self->decorateFromXML ($possibles[$#possibles], $instancechild);
            }
         }
      }
   }
}


# add a single value into the schema tree at a given position
sub addIntoTree
{
   my ($self, $schema, $name, $value, $upload) = @_;

   my $instancenumber = $name;
   $instancenumber =~ s/([^\.]*)\.([^\-])(.*)/$2/;

   # create more placeholders for new values if necessary
   my @values = $schema->getElementsByTagName ('value', 0);
   if ($instancenumber > ($#values+1))
   {
      for ( my $i=0; $i<($instancenumber-$#values-1); $i++ )
      { 
         my $clone;
         if (($schema->getAttribute ('type') eq 'string') ||
             ($schema->getElementsByTagName ('simpleType', 0)->getLength == 1))
         {
            my $parser = new Pure::EZXML;
            $clone = $parser->parse ('<value/>')->getDocumentElement;
            $schema->appendChild ($clone);
         }
         else
         {
            my $template = $schema->getElementsByTagName ('complexType', 0)->item(0);
            $clone = $template->cloneNode ('true');
            $clone->setTagName ('value');
            $schema->insertBefore ($clone, $template);
         }
         push (@values, $clone);
      }
   }
   
   my $instance = $values[$instancenumber-1];
   
   if (($schema->getAttribute ('type') eq 'string') ||
       ($schema->getElementsByTagName ('simpleType', 0)->getLength == 1))
   {
      while ($instance->hasChildNodes == 1)
      {
         $instance->removeChild ($instance->getFirstChild);
      }
      my $parser = new Pure::EZXML;
      if (($value ne '') && (($value ne '- none selected -') ||  ($schema->getAttribute ('type') eq 'string')))
      {
         # upload file and replace with URL if necessary
         if ($upload)
         {
            my $newvalue = $self->{filebox}->add ($value, $upload);
            close ($upload);
            if (defined $newvalue)
            {
               $value = $newvalue;
            }
         }
         
         my $frag = $parser->parse ("<t>$value</t>")->getDocumentElement;
         $instance->appendChild ($frag->getFirstChild);
      }
   }
   else
   {
      my $elementnumber = $name;
      $elementnumber =~ s/([^\.]*)\.([^\-])\-([^\.]*)(.*)/$3/;
      my @elements = $instance->getElementsByTagName ('sequence', 0)->item(0)
                              ->getElementsByTagName ('element', 0);
      my $newname = $name;
      $newname =~ s/([^\-]*)\-(.*)/$2/;
      $self->addIntoTree ($elements[$elementnumber-1], $newname, $value, $upload);
   }
}


# merge CGI variables with schema tree
sub decorateFromCGI
{
   my ($self, $schema, $cgi) = @_;
   
   if ($self->{cgideco} == 0)
   {
      $self->{cgideco} = 1;
   }
   else
   {
      return;
   }
   
   # find deleted file tag
   my $deletefile = undef;
   foreach my $name ($cgi->param)
   {
      if (substr ($name, 0, 11) eq 'deletefile.')
      {
         $deletefile = substr (substr ($name, 11), 0, -2);
      }
   }
   
   foreach my $name ($cgi->param)
   {
      if (substr ($name, 0, 2) eq '1.')
      {
         my @values = $cgi->param ($name);
         if ($name =~ /\-[0-9]+$/)
         {
            for ( my $i=0; $i<=$#values; $i++ )
            {
               $self->addIntoTree ($schema, $name.'.'.($i+1), $values[$i]);
            }
         }
         else
         {
            if ((defined $deletefile) && ($name eq $deletefile))
            {
               $self->{filebox}->delete ($values[0]);
               $values[0] = '';
            }
            $self->addIntoTree ($schema, $name, $values[0], $cgi->upload ($name));
         }
      }
   }
}


# add empty nodes into tree in order to input values that do not yet exist
sub completeTemplate
{
   my ($self, $schema) = @_;
   
   my $numberneeded = $schema->getAttribute ('minOccurs');
   if (($numberneeded eq '') || ($numberneeded == 0))
   {
      $numberneeded = 1;
   }
   
   if (($schema->getAttribute ('type') eq 'string') ||
       ($schema->getElementsByTagName ('simpleType', 0)->getLength == 1))
   {
      my @values = $schema->getElementsByTagName ('value', 0);
      for ( my $i=0; $i<($numberneeded-$#values-1); $i++ )
      {
         my $parser = new Pure::EZXML;
         my $frag = $parser->parse ('<value/>')->getDocumentElement;
         $schema->appendChild ($frag);
      }
   }
   else
   {
      my @newvalues = ();
      my @values = $schema->getElementsByTagName ('value', 0);
      for ( my $i=0; $i<($numberneeded-$#values-1); $i++ )
      {
         my $template = $schema->getElementsByTagName ('complexType', 0)->item(0);
         my $clone = $template->cloneNode ('true');
         $clone->setTagName ('value');
         $schema->appendChild ($clone);
         push (@newvalues, $clone);
      }
      @values = (@values, @newvalues);
      foreach my $value (@values)
      {
         foreach my $element ($value->getElementsByTagName ('sequence', 0)->item(0)
                                    ->getElementsByTagName ('element', 0))
         {
            $self->completeTemplate ($element);
         }
      }
   }
}


# name individual elements of the tree with unique hierarchical names
sub nameParts
{
   my ($self, $schema, $fieldname) = @_;
   
   if (! defined $fieldname)
   {
      $fieldname = '1';
   }
   
   $schema->setAttribute ('fieldname', $fieldname);
   
   my @values = $schema->getElementsByTagName ('value', 0);
   for ( my $i=0; $i<=$#values; $i++ )
   {
      my $valuefieldname = $fieldname.'.'.($i+1);
      $values[$i]->setAttribute ('fieldname', $valuefieldname);
      my @sequences = $values[$i]->getElementsByTagName ('sequence', 0);
      if ($#sequences == 0)
      {
         my @elements = $sequences[0]->getElementsByTagName ('element', 0);
         for ( my $j=0; $j<=$#elements; $j++ )
         {
            $self->nameParts ($elements[$j], $valuefieldname.'-'.($j+1));
         }
      }
   }
}


# get advanced option from schema description
sub getOption
{
   my ($self, $schema, $field, $default) = @_;
   
   my $annotation = $schema->getElementsByTagName ('annotation', 0);
   if ($annotation->getLength > 0)
   {
      my $appinfo = $annotation->item(0)->getElementsByTagName ('appinfo', 0);
      if ($appinfo->getLength > 0)
      {
         my $fieldtag = $appinfo->item(0)->getElementsByTagName ($field, 0);
         if ($fieldtag->getLength > 0)
         {
            if ($fieldtag->item(0)->hasChildNodes == 1)
            {
               return $fieldtag->item(0)->getChildNodes->toString;
            }
            else
            {
               return '';
            }
         }
      }
   }

   return $default;
}


# convert schema tree to a flattened list of html input fields
sub makeFormFields
{
   my ($self, $schema, $level) = @_;
   
   if (! defined $level)
   {
      $level = 2;
   }
   my $headingon = "<font face=\"arial\" size=\"".(6-$level)."\">";
   my $headingoff = "</font>";
   
   my $buffer = '';
   
   my $field_name = $schema->getAttribute ('name');
   my $field_label = $self->getOption ($schema, 'caption', $field_name);
   my $fixed_desc = $self->getOption ($schema, 'fixeddescription');
   my $is_a_file = $self->getOption ($schema, 'file');
   my $minOccurs = $schema->getAttribute ('minOccurs');

   my $numberneeded = $minOccurs;
   if (($numberneeded eq '') || ($numberneeded == 0))
   {
      $numberneeded = 1;
   }
   
   if ($schema->getAttribute ('type') eq 'string')
   {
      my $title = 0;
      my @values = $schema->getElementsByTagName ('value', 0);
      for ( my $i=0; $i<=$#values; $i++ )
      {
         my $value = '';
         if ($values[$i]->hasChildNodes == 1)
         {
            $value = $values[$i]->getFirstChild->getNodeValue;
         }

         $buffer .= '<tr>';
         if ($title == 0)
         {
            $buffer .= "<td ";
            if ($schema->getAttribute ('filled') eq 'false')
            {
               $buffer .= "bgcolor=\"$self->{missingcolor}\" ";
            }
            $buffer .= "valign=\"top\">$headingon$field_label$headingoff";
            if ($minOccurs > 0)
            {
               $buffer .= ' *';
            }
            if (defined $fixed_desc)
            {
               $buffer .= '<br>'.$fixed_desc;
            }
            $buffer .= "</td>";
            $title = 1;
         }
         else
         {
            $buffer .= '<td>&nbsp;</td>';
         }
         
         $buffer .= "<td valign=\"top\">";
         my $rows = $self->getOption ($schema, 'rows', 1);
         my $columns = $self->getOption ($schema, 'columns');
         my $fieldname = $values[$i]->getAttribute ('fieldname');
         my $inputtype = $self->getOption ($schema, 'inputtype', 'text');
         if (($inputtype eq 'file') && ($value ne ''))
         {
            if ($self->{filebox}->isLocal ($value))
            {
               $buffer .= "<a href=\"$value\">$value</a>".
                          " <input type=\"image\" name=\"deletefile.$fieldname\" src=\"$self->{delurl}\">".
                          "<input type=\"hidden\" name=\"$fieldname\" value=\"$value\">";
            }
            else
            {
               $buffer .= "<input name=\"$fieldname\" type=\"text\" value=\"$value\"";
               if (defined $columns)
               {
                  $buffer .= " size=\"$columns\"";
               }
               $buffer .= ">";
            }
         }
         else
         { 
            if ($rows > 1)
            {
               $buffer .= "<textarea name=\"$fieldname\" rows=\"$rows\" wrap=\"hard\"";
               if (defined $columns)
               {
                  $buffer .= " cols=\"$columns\"";
               }
               $buffer .= " wrap=\"off\">$value</textarea>";
            }
            else
            {
               $buffer .= "<input name=\"$fieldname\" type=\"$inputtype\" value=\"$value\"";
               if (defined $columns)
               {
                  $buffer .= " size=\"$columns\"";
               }
               $buffer .= ">";
            }
         }
         
         # add in buttons for additional fields if max has not been reached
         if ($i == $#values)
         {
            my $maxOccurs = $schema->getAttribute ('maxOccurs');
            if ($maxOccurs ne '')
            {
               if (($maxOccurs eq 'unbounded') || (($#values+1) < $maxOccurs))
               {
                  my $fieldname = $schema->getAttribute ('fieldname');
                  $buffer .= " <input type=\"image\" name=\"more.$fieldname\" src=\"$self->{plusurl}\">";
               }
            }
         }
         
         $buffer .= "</td></tr>\n";
      }
   }
   elsif ($schema->getElementsByTagName ('simpleType', 0)->getLength == 1)
   {
      $buffer .= "<td ";
      if ($schema->getAttribute ('filled') eq 'false')
      {
         $buffer .= "bgcolor=\"$self->{missingcolor}\" ";
      }
      $buffer .= "valign=\"top\">$headingon$field_label$headingoff";
      if ($minOccurs > 0)
      {
         $buffer .= ' *';
      }
      if (defined $fixed_desc)
      {
         $buffer .= '<br>'.$fixed_desc;
      }
      $buffer .= "</td>";
      
      my %selected = ();
      my $gotselected = 0;
      foreach my $value ($schema->getElementsByTagName ('value', 0))
      {
         if ($value->hasChildNodes == 1)
         {
            $selected{$value->getFirstChild->getNodeValue} = '1';
            $gotselected = 1;
         }
      }
      
      $buffer .= "<td valign=\"top\">";
      my $rows = $self->getOption ($schema, 'rows', 1);
      my $fieldname = $schema->getAttribute ('fieldname');
      my $inputtype = $self->getOption ($schema, 'inputtype', 'select');
      if ($inputtype eq 'radio')
      {
         my $nonechecked = ($gotselected == 1) ? '' : ' checked';
         $buffer .= "<input name=\"$fieldname\" type=\"radio\" value=\"\"$nonechecked>(none)<br>";
         foreach my $simpletype ($schema->getElementsByTagName ('simpleType', 0))
         {
            foreach my $restriction ($simpletype->getElementsByTagName ('restriction', 0))
            {
               my @enumerations = $restriction->getElementsByTagName ('enumeration', 0);
               if ($#enumerations > -1)
               {
                  foreach my $enumeration (@enumerations)
                  {
                     my $enumvalue = $enumeration->getAttribute ('value');
                     $buffer .= "<input name=\"$fieldname\" type=\"radio\" value=\"$enumvalue\"";
                     if ($selected{$enumvalue} == 1)
                     {
                        $buffer .= ' checked';
                     }
                     $buffer .= '>';
                     $buffer .= $enumvalue;
                     $buffer .= '<br>';
                  }
               }
               else
               {
                  my @selectedkeys = keys %selected;
                  if ($#selectedkeys > -1)
                  {
                     $buffer .= "<input type=\"radio\" name=\"$fieldname\" value=\"$selectedkeys[0]\" checked>$selectedkeys[0]<br>";
                  }
               }
            }
         }
      }
      else
      {
         foreach my $simpletype ($schema->getElementsByTagName ('simpleType', 0))
         {
            foreach my $restriction ($simpletype->getElementsByTagName ('restriction', 0))
            {
               my @enumerations = $restriction->getElementsByTagName ('enumeration', 0);

               $buffer .= "<select name=\"$fieldname\" size=\"$rows\"";
               my $firstoption = '';
               if (($schema->getAttribute ('maxOccurs') == 'unbounded') ||
                   ($schema->getAttribute ('maxOccurs') > 1))
               {
                  $buffer .= ' multiple';
               }
               else
               {
                  $firstoption = '<option>- none selected -</option>';
               }
               $buffer .= '>';

               if ($#enumerations > -1)
               {
                  $buffer .= $firstoption;
                  foreach my $enumeration (@enumerations)
                  {
                     my $enumvalue = $enumeration->getAttribute ('value');
                     $buffer .= '<option';
                     if ($selected{$enumvalue} == 1)
                     {
                        $buffer .= ' selected';
                     }
                     $buffer .= '>';
                     $buffer .= $enumvalue;
                     $buffer .= '</option>';
                  }
               }
               else
               {
                  foreach my $enumeration (keys %selected)
                  {
                     $buffer .= "<option selected>$enumeration</option>";
                  }
               }
               
               $buffer .= '</select>';
            }
         }
      }
      $buffer .= "</td></tr>\n";
   }
   else
   { 
      my $more;
      my @values = $schema->getElementsByTagName ('value', 0);

      # add in buttons for additional fields if max has not been reached
      my $maxOccurs = $schema->getAttribute ('maxOccurs');
      if ($maxOccurs ne '')
      {
         if (($maxOccurs eq 'unbounded') || (($#values+1) < $maxOccurs))
         {
            my $fieldname = $schema->getAttribute ('fieldname');
            $more = " <input type=\"image\" name=\"more.$fieldname\" src=\"$self->{plusurl}\">";
         }
      }
      
      for ( my $i=0; $i<=$#values; $i++ )
      {
         if ($field_label ne '')
         {
            $buffer .= "<tr><td>&nbsp;</td></tr>".
                       "<tr><td";
            if ($schema->getAttribute ('filled') eq 'false')
            {
               $buffer .= " bgcolor=\"$self->{missingcolor}\"";
            }                       
            $buffer .= ">$headingon$field_label";
            if (($minOccurs > 0) && ($i < $minOccurs)) 
            {
               $buffer .= ' *';
            }
            if ($i == $#values)
            {
               $buffer .= $more;
            }
            $buffer .= $headingoff;
            if (defined $fixed_desc)
            {
               $buffer .= '<br>'.$fixed_desc;
            }
            $buffer .= "</td></tr>";
         }
         foreach my $element ($values[$i]->getElementsByTagName ('sequence', 0)->item(0)
                                         ->getElementsByTagName ('element', 0))
         {
            $buffer .= $self->makeFormFields ($element, $level+1);
         }
      }
      
      $buffer .= '<tr><td>&nbsp;</td></tr>';
   }
   
   return $buffer;
}


# create an HTML form based on the schema
sub makeForm
{
   my ($self, $schema, $embed) = @_;
   
   my $elements = $self->makeFormFields ($schema);
   
   $elements =~ s/&amp;(?=nbsp;)/&/go;
   
   my $embedstr = '';
   if (defined $embed)
   {
      foreach my $key (keys %$embed)
      {
         $embedstr .= "<input type=\"hidden\" name=\"$key\" ".
                      "value=\"$embed->{$key}\">";
      }
   }
   
   "<form method=\"post\" enctype=\"multipart/form-data\" multipart/form-data><table border=\"0\" cellpadding=\"3\">".
   $elements.
   "</table><input type=\"submit\" name=\"submit\" value=\"Submit\">".
   "$embedstr</form>";
}


# add an additional entry field in response to a "more" request
sub addField
{
   my ($self, $schema, $fieldspec) = @_;
   
   if (index ($fieldspec, '.') == -1)
   {
      # no dots in fieldspec => make copy
      # ignore simpleType since they are not replicated ?
      if ($schema->getAttribute ('type') ne 'string')
      {
         my $template = $schema->getElementsByTagName ('complexType', 0)->item(0);
         my $clone = $template->cloneNode ('true');
         $clone->setTagName ('value');
         $schema->insertBefore ($clone, $template);
      }
      else
      {
         my $parser = new Pure::EZXML;
         my $clone = $parser->parse ('<value/>')->getDocumentElement;
         $schema->appendChild ($clone);
      }
   }
   else
   {
      my $instancenumber = $fieldspec;
      $instancenumber =~ s/([^\.]*)\.([^\-])(.*)/$2/;
      
      my @values = $schema->getElementsByTagName ('value', 0);
      my $instance = $values[$instancenumber-1];
      
      my $elementnumber = $fieldspec;
      $elementnumber =~ s/([^\.]*)\.([^\-])\-([^\.]*)(.*)/$3/;
      
      my @elements = $instance->getElementsByTagName ('sequence', 0)->item(0)
                              ->getElementsByTagName ('element', 0);
      my $newname = $fieldspec;
      $newname =~ s/([^\-]*)\-(.*)/$2/;
      $self->addField ($elements[$elementnumber-1], $newname);
   }
}


# convert schema tree with data into XML data representation
sub makeXMLFields
{
   my ($self, $schema, $indent) = @_;
   
   if (! defined $indent)
   {
      $indent = '';
   }
   else
   {
      $indent .= '   ';
   }
   
   my $buffer = '';
   
   foreach my $value ($schema->getElementsByTagName ('value', 0))
   {
      my $nsschema = '';
      if ($schema == $self->{schema})
      {
         my $targetns = $schema->getDocumentElement->getAttribute ('targetNamespace');
         $nsschema = " xmlns=\"$targetns\"";
         if (exists $self->{schemaURL})
         {
            $nsschema .= " xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"".
                         " xsi:schemaLocation=\"$targetns $self->{schemaURL}\"";
         }
      }
      if (($schema->getAttribute ('type') eq 'string') ||
          ($schema->getElementsByTagName ('simpleType', 0)->getLength == 1))
      {
         if ($value->hasChildNodes == 1)
         {
            my $valuestr = $value->getFirstChild->getNodeValue;
            if ($valuestr !~ /^[\s\n\r\t]+$/)
            {
               $buffer .= $indent.'<'.$schema->getAttribute ('name').$nsschema.'>'.
                          $valuestr.
                          '</'.$schema->getAttribute ('name').'>'."\n";
            }
         }
      }
      else
      {
         my $childbuffer = '';
         foreach my $element ($value->getElementsByTagName ('sequence', 0)->item(0)
                                    ->getElementsByTagName ('element', 0))
         {
            $childbuffer .= $self->makeXMLFields ($element, $indent);
         }
         if ($childbuffer ne '')
         {
            $buffer .= $indent.'<'.$schema->getAttribute ('name').$nsschema.'>'."\n";
            $buffer .= $childbuffer;
            $buffer .= $indent.'</'.$schema->getAttribute ('name').'>'."\n";
         }
      }
   }
   
   return $buffer;
}


# make XML document
sub makeXML
{
   my ($self, $schema) = @_;
   
   my $elements = $self->makeXMLFields ($schema);
   
   return $elements;
}


# check if any of the CGI parameters indicate a field needs to be duplicated
sub getMore 
{
   my ($self, $cgi) = @_;
   
   foreach my $name ($cgi->param)
   {
      if (substr ($name, 0, 5) eq 'more.')
      {
         return substr (substr ($name, 5), 0, -2);
      }
   }
   
   return '';
}


# create HTML form from a schema (and instance)
sub formFromSchema
{
   my ($self, $embed, $instance) = @_;
   if (defined $instance)
   {
      $self->decorateFromXML ($self->{schema}, $instance);
   }
   $self->completeTemplate ($self->{schema});
   $self->nameParts ($self->{schema});
   return $self->makeForm ($self->{schema}, $embed);
}


# create HTML form from CGI variables
sub formFromCGI
{
   my ($self, $cgi, $embed) = @_;
   
   $self->decorateFromCGI ($self->{schema}, $cgi);
   my $morefield = $self->getMore ($cgi);
   if ($morefield ne '')
   {
      $self->addField ($self->{schema}, $morefield);
   }
   $self->completeTemplate ($self->{schema});
   $self->nameParts ($self->{schema});
   return $self->makeForm ($self->{schema}, $embed);
}


# create XML from CGI variables
sub XMLFromCGI
{
   my ($self, $cgi) = @_;
   
   $self->decorateFromCGI ($self->{schema}, $cgi);
   $self->completeTemplate ($self->{schema});
   $self->nameParts ($self->{schema});
   return $self->makeXML ($self->{schema});
}


# remove empty and unused nodes from the tree
sub prune
{
   my ($self, $schema) = @_;
   
   my $buffer = '';
   
   foreach my $value ($schema->getElementsByTagName ('value', 0))
   {
      if (($schema->getAttribute ('type') eq 'string') ||
          ($schema->getElementsByTagName ('simpleType', 0)->getLength == 1))
      {
         if ($value->hasChildNodes == 1)
         {
            my $valuestr = $value->getFirstChild->getNodeValue;
            if ($valuestr !~ /^[\s\n\r\t]+$/)
            {
               $buffer .= '1';
            }
            else
            {
               $value->getParentNode->removeChild ($value);
            }
         }
         else
         {
            $value->getParentNode->removeChild ($value);
         }
      }
      else
      {
         my $childbuffer = '';
         foreach my $element ($value->getElementsByTagName ('sequence', 0)->item(0)
                                    ->getElementsByTagName ('element', 0))
         {
            $childbuffer .= $self->prune ($element);
         }
         if ($childbuffer ne '')
         {
            $buffer .= '1';
         }
         else
         {
            $value->getParentNode->removeChild ($value);
         }
      }
   }
   
   return $buffer;
}


# count the number of missing fields in the schema tree
sub sumMissing
{
   my ($self, $schema) = @_;

   my $checks = 0;
   my $minOccurs = $schema->getAttribute ('minOccurs');
   if ($minOccurs eq '')
   {
      $minOccurs = 0;
   }
   my @values = $schema->getElementsByTagName ('value', 0);
   if (($schema->getAttribute ('type') eq 'string') ||
       ($schema->getElementsByTagName ('simpleType', 0)->getLength == 1))
   {
      if (($#values+1) < $minOccurs)
      {
         $schema->setAttribute ('filled', 'false');
         $checks += 1;
      }
   }
   else
   {
      if (($#values+1) < $minOccurs)
      {
         $schema->setAttribute ('filled', 'false');
         $checks += 1;
      }
      foreach my $value (@values)
      {
         foreach my $element ($value->getElementsByTagName ('sequence', 0)->item(0)
                                    ->getElementsByTagName ('element', 0))
         {
            $checks += $self->sumMissing ($element);
         }
      }
   }
   return $checks;
}


# check if the schema satisfies the min/max conditions
sub check
{
   my ($self, $cgi) = @_;
   
   $self->decorateFromCGI ($self->{schema}, $cgi);
   $self->prune ($self->{schema});
   
   return $self->sumMissing ($self->{schema});
}

1;

